@@ -1,9 +1,183 @@
 """ Class to work with Salesforce Metadata API """
+import os
 from base64 import b64encode, b64decode
 from xml.etree import ElementTree as ET
 from .util import call_salesforce
-from .messages import DEPLOY_MSG,CHECK_DEPLOY_STATUS_MSG,\
-    CHECK_RETRIEVE_STATUS_MSG,RETRIEVE_MSG
+from .messages import DEPLOY_MSG, CHECK_DEPLOY_STATUS_MSG, \
+    CHECK_RETRIEVE_STATUS_MSG, RETRIEVE_MSG
+from zeep import Client, Settings
+
+
+class MetadataType:
+    """
+    Salesforce Metadata Type
+    """
+    def __init__(self, name, service, zeep_type, session_header):
+        """
+        Initialize metadata type
+
+        :param name: Name of metadata type
+        :type name: str
+        :param service: Zeep service
+        :type service: zeep.proxy.ServiceProxy
+        :param zeep_type: Zeep type object
+        :type zeep_type: zeep.xsd.ComplexType or zeep.xsd.AnySimpleType
+        :param session_header: Session Id header for Metadata API calls
+        """
+        self._name = name
+        self._service = service
+        self._zeep_type = zeep_type
+        self._session_header = session_header
+
+    @staticmethod
+    def _handle_api_response(response):
+        """
+        Parses SaveResult and DeleteResult objects to identify if there was
+        an error, and raises exception accordingly
+
+        :param response: List of zeep.objects.SaveResult or
+        zeep.objects.DeleteResult objects
+        :type response: list
+        :raises Exception: If any Result object contains one or more error
+        messages
+        """
+        err_string = ""
+        for result in response:
+            if not result.success:
+                err_string += "\n{}: ".format(result.fullName)
+                for error in result.errors:
+                    err_string += "({}, {}), ".format(error.statusCode,
+                                                      error.message)
+        if err_string:
+            raise Exception(err_string)
+
+    def __call__(self, *args, **kwargs):
+        """
+        Creates a new object of this metadata type
+
+        :param args: Parameters to pass to zeep.xsd.AnySimpleType
+        :param kwargs: Parameters to pass to zeep.xsd.ComplexType
+        :returns: An object of type self._name
+        """
+        return self._zeep_type(*args, **kwargs)
+
+    def create(self, metadata):
+        """
+        Performs a createMetadata call
+
+        :param metadata: Array of one or more metadata components.
+                         Limit: 10. (For CustomMetadata and CustomApplication
+                         only, the limit is 200.)
+                         You must submit arrays of only one type of
+                         component. For example, you can submit an
+                         array of 10 custom objects or 10 profiles, but not a
+                         mix of both types.
+        :type metadata: list
+        """
+        response = self._service.createMetadata(metadata, _soapheaders=[
+            self._session_header])
+        self._handle_api_response(response)
+
+    def read(self, full_names):
+        """
+        Performs a readMetadata call
+
+        :param full_names: Array of full names of the components to read.
+                           Limit: 10. (For CustomMetadata and
+                           CustomApplication only, the limit is 200.)
+                           You must submit arrays of only one type of
+                           component. For example, you can submit an array
+                           of 10 custom objects or 10 profiles, but not a mix
+                           of both types.
+        :type full_names: list
+        :returns: A list of metadata components
+        :rtype: list
+        """
+        response = self._service.readMetadata(self._name, full_names,
+                                              _soapheaders=[
+                                                  self._session_header])
+        if len(response) == 1:
+            return response[0]
+        return response
+
+    def update(self, metadata):
+        """
+        Performs an updateMetadata call. All required fields must be passed
+        for each component
+
+        :param metadata: Array of one or more metadata components.
+                         Limit: 10. (For CustomMetadata and CustomApplication
+                         only, the limit is 200.)
+                         You must submit arrays of only one type of
+                         component. For example, you can submit an
+                         array of 10 custom objects or 10 profiles, but not a
+                         mix of both types.
+        :type metadata: list
+        """
+        response = self._service.updateMetadata(metadata, _soapheaders=[
+            self._session_header])
+        self._handle_api_response(response)
+
+    def upsert(self, metadata):
+        """
+        Performs an upsertMetadata call. All required fields must be passed
+        for each component
+
+        :param metadata: Array of one or more metadata components.
+                         Limit: 10. (For CustomMetadata and CustomApplication
+                         only, the limit is 200.)
+                         You must submit arrays of only one type of
+                         component. For example, you can submit an
+                         array of 10 custom objects or 10 profiles, but not a
+                         mix of both types.
+        :type metadata: list
+        """
+        response = self._service.updateMetadata(metadata, _soapheaders=[
+            self._session_header])
+        self._handle_api_response(response)
+
+    def delete(self, full_names):
+        """
+        Performs a deleteMetadata call
+
+        :param full_names: Array of full names of the components to delete.
+                           Limit: 10. (For CustomMetadata and
+                           CustomApplication only, the limit is 200.)
+                           You must submit arrays of only one type of
+                           component. For example, you can submit an array
+                           of 10 custom objects or 10 profiles, but not a mix
+                           of both types.
+        :type full_names: list
+        """
+        response = self._service.deleteMetadata(self._name, full_names,
+                                                _soapheaders=[
+                                                    self._session_header])
+        self._handle_api_response(response)
+
+    def rename(self, old_full_name, new_full_name):
+        """
+        Performs a renameMetadata call
+
+        :param old_full_name: The current component full name.
+        :type old_full_name: str
+        :param new_full_name: The new component full name.
+        :type new_full_name: str
+        """
+        result = self._service.renameMetadata(self._name, old_full_name,
+                                              new_full_name,
+                                              _soapheaders=[
+                                                  self._session_header])
+        self._handle_api_response([result])
+
+    def describe(self):
+        """
+        Performs a describeValueType call
+
+        :returns: DescribeValueTypeResult
+        """
+        return self._service.describeValueType(
+            "{{http://soap.sforce.com/2006/04/metadata}}{}".format(self._name),
+            _soapheaders=[self._session_header])
 
 
 class SfdcMetadataApi:
@@ -16,21 +190,58 @@ class SfdcMetadataApi:
         }
 
     # pylint: disable=R0913
-    def __init__(self, session, session_id, instance, sandbox, metadata_url,
-                 headers, api_version):
+    def __init__(self, session, session_id, instance, metadata_url, headers,
+                 api_version):
         """ Initialize and check session """
         self.session = session
         self._session_id = session_id
         self._instance = instance
-        self._sandbox = sandbox
         self.metadata_url = metadata_url
         self.headers = headers
         self._api_version = api_version
         self._deploy_zip = None
+        wsdl_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
+                                 'metadata.wsdl')
+        self._client = Client(os.path.join('simple_salesforce', wsdl_path),
+                              settings=Settings(strict=False,
+                                                xsd_ignore_sequence_order=True))
+        self._service = self._client.create_service(
+            "{http://soap.sforce.com/2006/04/metadata}MetadataBinding",
+            self.metadata_url)
+        self._session_header = self._client.get_element('ns0:SessionHeader')(
+            sessionId=self._session_id)
+
+    def __getattr__(self, item):
+        return MetadataType(item, self._service,
+                            self._client.get_type('ns0:' + item),
+                            self._session_header)
+
+    def describe_metadata(self):
+        """
+        Performs a describeMetadata call
+
+        :returns: An object of zeep.objects.DescribeMetadataResult
+        """
+        return self._service.describeMetadata(self._api_version, _soapheaders=[
+            self._session_header])
+
+    def list_metadata(self, queries):
+        """
+        Performs a listMetadata call
+
+        :param queries: A list of zeep.objects.ListMetadataQuery that specify
+        which components you are interested in.
+                        Limit: 3
+        :type queries: list
+        :returns: List of zeep.objects.FileProperties objects
+        :rtype: list
+        """
+        return self._service.listMetadata(queries, self._api_version,
+                                          _soapheaders=[self._session_header])
 
     # pylint: disable=R0914
     # pylint: disable-msg=C0103
-    def deploy(self, zipfile, **kwargs):
+    def deploy(self, zipfile, sandbox, **kwargs):
         """ Kicks off async deployment, returns deployment id
         :param zipfile:
         :type zipfile:
@@ -67,7 +278,7 @@ class SfdcMetadataApi:
             'singlePackage': singlePackage,
             }
 
-        if not self._sandbox:
+        if not sandbox:
             attributes['allowMissingFiles'] = False
             attributes['rollbackOnError'] = True
 
@@ -77,7 +288,7 @@ class SfdcMetadataApi:
 
         tests_tag = ''
         if tests and \
-                str(tests).lower() == 'runspecifiedtests':
+                str(testLevel).lower() == 'runspecifiedtests':
             for test in tests:
                 tests_tag += '<met:runTests>%s</met:runTests>\n' % test
             attributes['tests'] = tests_tag
@@ -245,8 +456,8 @@ class SfdcMetadataApi:
         """ Downloads Apex logs for unit tests executed during specified
         deployment """
         result = self._retrieve_deploy_result(async_process_id)
-        print("Results: %s" % ET.tostring(result, encoding="us-ascii",
-                                          method="xml"))
+        print("response: %s" % ET.tostring(result, encoding="us-ascii",
+                                           method="xml"))
 
     def retrieve(self, async_process_id, **kwargs):
         """ Submits retrieve request """
@@ -291,7 +502,7 @@ class SfdcMetadataApi:
             additional_headers=headers,
             data=request)
 
-        # Parse results to get async Id and status
+        # Parse response to get async Id and status
         async_process_id = ET.fromstring(res.text).find(
             'soapenv:Body/mt:retrieveResponse/mt:result/mt:id',
             self._XML_NAMESPACES).text
@@ -334,7 +545,7 @@ class SfdcMetadataApi:
     def retrieve_zip(self, async_process_id, **kwargs):
         """ Retrieves ZIP file """
         result = self.retrieve_retrieve_result(async_process_id, 'true',
-                                                **kwargs)
+                                               **kwargs)
         state = result.find('mt:status', self._XML_NAMESPACES).text
         error_message = result.find('mt:errorMessage', self._XML_NAMESPACES)
         if error_message is not None:
@@ -359,7 +570,7 @@ class SfdcMetadataApi:
     def check_retrieve_status(self, async_process_id, **kwargs):
         """ Checks whether retrieval succeeded """
         result = self.retrieve_retrieve_result(async_process_id, 'false',
-                                                **kwargs)
+                                               **kwargs)
         state = result.find('mt:status', self._XML_NAMESPACES).text
         error_message = result.find('mt:errorMessage', self._XML_NAMESPACES)
         if error_message is not None:
